@@ -1,17 +1,26 @@
+//! A milter that prints callback arguments and macros for each stage.
+
 use indymilter::{
     Actions, Callbacks, Context, EomContext, Macros, NegotiateContext, ProtoOpts, SocketInfo,
     Status,
 };
-use std::ffi::CString;
+use std::{env, ffi::CString, process};
 use tokio::{net::TcpListener, signal};
 
-// TODO or use tracing?
-
 #[tokio::main]
 async fn main() {
     tracing_subscriber::fmt::init();
 
-    let config = Default::default();
+    let args = env::args().collect::<Vec<_>>();
+
+    if args.len() != 2 {
+        eprintln!("usage: {} <socket>", args[0]);
+        process::exit(1);
+    }
+
+    let listener = TcpListener::bind(&args[1])
+        .await
+        .expect("cannot open milter socket");
 
     let callbacks = Callbacks::new()
         .on_negotiate(|cx, actions, opts| Box::pin(handle_negotiate(cx, actions, opts)))
@@ -28,9 +37,7 @@ async fn main() {
         .on_close(|cx| Box::pin(handle_close(cx)))
         .on_unknown(|cx, arg| Box::pin(handle_unknown(cx, arg)));
 
-    let listener = TcpListener::bind("localhost:3000")
-        .await
-        .expect("cannot open milter socket");
+    let config = Default::default();
 
     indymilter::run(listener, callbacks, config, signal::ctrl_c())
         .await
@@ -52,36 +59,36 @@ async fn handle_negotiate(
 async fn handle_connect(
     cx: &mut Context<()>,
     hostname: CString,
-    socket_info: Option<SocketInfo>,
+    socket_info: SocketInfo,
 ) -> Status {
     println!("CONNECT");
-    print_macros(&cx.macros);
     println!("  hostname: {:?}", hostname);
     println!("  socket_info: {:?}", socket_info);
+    print_macros(&cx.macros);
 
     Status::Continue
 }
 
 async fn handle_helo(cx: &mut Context<()>, hostname: CString) -> Status {
     println!("HELO");
-    print_macros(&cx.macros);
     println!("  hostname: {:?}", hostname);
+    print_macros(&cx.macros);
 
     Status::Continue
 }
 
 async fn handle_mail(cx: &mut Context<()>, args: Vec<CString>) -> Status {
     println!("MAIL");
-    print_macros(&cx.macros);
     println!("  args: {:?}", args);
+    print_macros(&cx.macros);
 
     Status::Continue
 }
 
 async fn handle_rcpt(cx: &mut Context<()>, args: Vec<CString>) -> Status {
     println!("RCPT");
-    print_macros(&cx.macros);
     println!("  args: {:?}", args);
+    print_macros(&cx.macros);
 
     Status::Continue
 }
