@@ -17,41 +17,56 @@ pub enum Status {
     AllOpts,
 }
 
+/// The type returned by milter callbacks.
 pub type CallbackFuture<'a> = Pin<Box<dyn Future<Output = Status> + Send + 'a>>;
 
 // Default trait enables struct update syntax (`..Default::default()`)
 /// Callback closures for each milter stage.
 #[derive(Default)]
 pub struct Callbacks<T: Send> {
+    /// The callback for the `negotiate` stage.
     pub negotiate:
         Option<Box<dyn for<'cx> Fn(&'cx mut NegotiateContext<T>, Actions, ProtoOpts) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `connect` stage.
     pub connect:
-        Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, CString, Option<SocketInfo>) -> CallbackFuture<'cx> + Send + Sync>>,
+        Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, CString, SocketInfo) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `helo` stage.
     pub helo:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, CString) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `mail` stage.
     pub mail:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, Vec<CString>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `rcpt` stage.
     pub rcpt:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, Vec<CString>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `data` stage.
     pub data:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `header` stage.
     pub header:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, CString, CString) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `eoh` stage.
     pub eoh:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `body` stage.
     pub body:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, Vec<u8>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `eom` stage.
     pub eom:
         Option<Box<dyn for<'cx> Fn(&'cx mut EomContext<T>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `abort` stage.
     pub abort:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `close` stage.
     pub close:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>) -> CallbackFuture<'cx> + Send + Sync>>,
+    /// The callback for the `unknown` stage.
     pub unknown:
         Option<Box<dyn for<'cx> Fn(&'cx mut Context<T>, CString) -> CallbackFuture<'cx> + Send + Sync>>,
 }
 
 impl<T: Send> Callbacks<T> {
+    /// Returns a new, empty collection of callbacks.
     pub fn new() -> Self {
         Self {
             negotiate: None,
@@ -80,7 +95,7 @@ impl<T: Send> Callbacks<T> {
 
     pub fn on_connect(
         mut self,
-        callback: impl Fn(&mut Context<T>, CString, Option<SocketInfo>) -> CallbackFuture<'_> + Send + Sync + 'static,
+        callback: impl Fn(&mut Context<T>, CString, SocketInfo) -> CallbackFuture<'_> + Send + Sync + 'static,
     ) -> Self {
         self.connect = Some(Box::new(callback));
         self
