@@ -6,6 +6,7 @@ use crate::{
     proto_util::{Actions, ProtoOpts},
 };
 use async_trait::async_trait;
+use bytes::Bytes;
 use std::{
     collections::HashMap,
     error::Error,
@@ -15,26 +16,51 @@ use std::{
     str::FromStr,
 };
 
+/// Customisation of an SMTP reply.
 pub trait SetErrorReply {
+    /// Configures an SMTP error reply.
     fn set_error_reply<I, T>(
         &mut self,
         rcode: &str,
         xcode: Option<&str>,
         message: I,
-    ) -> Result<(), ContextError>
+    ) -> Result<(), SmtpReplyError>
     where
         I: IntoIterator<Item = T>,
         T: IntoCString;
 }
 
+/// An error that occurs when configuring the SMTP reply.
+#[derive(Debug)]
+pub enum SmtpReplyError {
+    InvalidReplyCode,
+    InvalidEnhancedStatusCode,
+    InvalidReplyText,
+}
+
+impl Display for SmtpReplyError {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::InvalidReplyCode => write!(f, "invalid SMTP reply code"),
+            Self::InvalidEnhancedStatusCode => write!(f, "invalid enhanced status code"),
+            Self::InvalidReplyText => write!(f, "invalid SMTP reply text"),
+        }
+    }
+}
+
+impl Error for SmtpReplyError {}
+
+/// Context actions.
 #[async_trait]
 pub trait ContextActions {
+    /// Adds a header.
     async fn add_header<'cx, 'k, 'v>(
         &'cx self,
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
     ) -> Result<(), ContextError>;
 
+    /// Inserts a header at the given index.
     async fn insert_header<'cx, 'k, 'v>(
         &'cx self,
         index: i32,
@@ -42,6 +68,7 @@ pub trait ContextActions {
         value: impl IntoCString + Send + 'v,
     ) -> Result<(), ContextError>;
 
+    /// Replaces or deletes the header at the given index.
     async fn change_header<'cx, 'k, 'v>(
         &'cx self,
         name: impl IntoCString + Send + 'k,
@@ -49,57 +76,72 @@ pub trait ContextActions {
         value: Option<impl IntoCString + Send + 'v>,
     ) -> Result<(), ContextError>;
 
+    /// Replaces the envelope sender, with optional ESMTP arguments.
     async fn change_sender<'cx, 'a, 'b>(
         &'cx self,
         mail: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
     ) -> Result<(), ContextError>;
 
+    /// Adds an envelope recipient.
     async fn add_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
     ) -> Result<(), ContextError>;
 
+    /// Adds an envelope recipient, with ESMTP arguments.
     async fn add_recipient_ext<'cx, 'a, 'b>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
     ) -> Result<(), ContextError>;
 
+    /// Removes an envelope recipient.
     async fn delete_recipient<'cx, 'a>(
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
     ) -> Result<(), ContextError>;
 
+    /// Replaces the message body.
     async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ContextError>;
 
+    /// Reports operation-in-progress to the MTA.
     async fn progress<'cx>(&'cx self) -> Result<(), ContextError>;
 
+    /// Quarantines the message.
     async fn quarantine<'cx, 'a>(
         &'cx self,
         reason: impl IntoCString + Send + 'a,
     ) -> Result<(), ContextError>;
 }
 
+// TODO Rename ActionError?
+/// An error that occurs when executing a context action.
 #[derive(Debug)]
 pub enum ContextError {
     ActionUnavailable,
-    InvalidReply,
+    InvalidParams,
     Io(io::Error),
-
-    // TODO separate out into separate type:
-    InvalidReplyCode,
-    InvalidEnhancedStatusCode,
-    InvalidReplyText,
 }
 
 impl Display for ContextError {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        write!(f, "{:?}", self)
+        match self {
+            Self::ActionUnavailable => write!(f, "action has not been enabled"),
+            Self::InvalidParams => write!(f, "invalid context action parameters"),
+            Self::Io(error) => write!(f, "I/O error: {}", error),
+        }
     }
 }
 
-impl Error for ContextError {}
+impl Error for ContextError {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
+        match self {
+            Self::Io(error) => Some(error),
+            _ => None,
+        }
+    }
+}
 
 impl From<io::Error> for ContextError {
     fn from(error: io::Error) -> Self {
@@ -109,10 +151,15 @@ impl From<io::Error> for ContextError {
 
 /// Callback context for the `negotiate` stage.
 pub struct NegotiateContext<T: Send> {
+    /// The user data.
     pub data: Option<T>,
+    /// The custom SMTP reply.
     pub reply: SmtpReply,
+    /// The actions to request from the MTA.
     pub requested_actions: Actions,
+    /// The milter protocol options to request from the MTA.
     pub requested_opts: ProtoOpts,
+    /// The macros to request from the MTA.
     pub requested_macros: HashMap<Stage, CString>,
 }
 
@@ -133,13 +180,19 @@ impl<T: Send> NegotiateContext<T> {
     }
 }
 
+// With the current design, callbacks receive `&mut` references to context
+// structs. Thus they have mutable access to everything in the struct. However,
+// this means that callbacks have the ability to extract eg `Macros` and
+// `SmtpReply`. To prevent this, these structs cannot be instantiated by users
+// and do not implement `Default` or `Clone`.
+
 /// Callback context.
 pub struct Context<T: Send> {
+    /// The user data.
     pub data: Option<T>,
-
-    // TODO if we provide a &mut reference to the context then this is all open,
-    // and can be replaced by users; must somehow be read-only/protected/uninstantiable
+    /// The defined macros.
     pub macros: Macros,
+    /// The custom SMTP reply.
     pub reply: SmtpReply,
 }
 
@@ -169,26 +222,30 @@ impl<T: Send> Context<T> {
         self.reply = reply;
     }
 
-    pub(crate) fn take_reply_if<F>(&mut self, f: F) -> Option<Reply>
+    pub(crate) fn take_reply_if<F>(&mut self, predicate: F) -> Option<Reply>
     where
         F: FnMut(&ReplyCode) -> bool,
     {
         self.reply
-            .take_reply_if_match(f)
+            .take_reply_if_match(predicate)
             .map(|reply| Reply::ReplyCode { reply })
     }
 }
 
 /// Callback context for the `eom` stage.
 pub struct EomContext<T: Send> {
+    /// The user data.
     pub data: Option<T>,
+    /// The defined macros.
     pub macros: Macros,
+    /// The custom SMTP reply.
     pub reply: SmtpReply,
+    /// The end-of-message actions.
     pub actions: EomActions,
 }
 
 impl<T: Send> EomContext<T> {
-    pub fn new(
+    pub(crate) fn new(
         conn: Connection,
         data: Option<T>,
         macros: Macros,
@@ -207,7 +264,8 @@ impl<T: Send> EomContext<T> {
     }
 }
 
-// No Default, Clone, must be uninstantiable by users.
+// Note: This struct does *not* implement `Default` or `Clone`. Callbacks
+// receive a mutable reference to this struct but must not be able to remove it.
 /// Current custom SMTP reply.
 #[derive(Debug)]
 pub struct SmtpReply {
@@ -226,12 +284,12 @@ impl SmtpReply {
     }
 
     // TODO revisit
-    pub(crate) fn take_reply_if_match<F>(&mut self, mut f: F) -> Option<CString>
+    pub(crate) fn take_reply_if_match<F>(&mut self, mut predicate: F) -> Option<CString>
     where
         F: FnMut(&ReplyCode) -> bool,
     {
         if let Some(reply) = self.error_reply.as_ref() {
-            if f(&reply.rcode) {
+            if predicate(&reply.rcode) {
                 return self.error_reply.take().map(|r| r.make_error_reply());
             }
         }
@@ -245,32 +303,34 @@ impl SetErrorReply for SmtpReply {
         rcode: &str,
         xcode: Option<&str>,
         message: I,
-    ) -> Result<(), ContextError>
+    ) -> Result<(), SmtpReplyError>
     where
         I: IntoIterator<Item = T>,
         T: IntoCString,
     {
-        let rcode = rcode.parse().map_err(|_| ContextError::InvalidReplyCode)?;
+        let rcode = rcode
+            .parse()
+            .map_err(|_| SmtpReplyError::InvalidReplyCode)?;
 
         let xcode = xcode
             .map(str::parse)
             .transpose()
-            .map_err(|_| ContextError::InvalidEnhancedStatusCode)?;
+            .map_err(|_| SmtpReplyError::InvalidEnhancedStatusCode)?;
 
         let mut message_lines = Vec::new();
 
         for (i, line) in message.into_iter().enumerate() {
             if i >= 32 {
-                return Err(ContextError::InvalidReplyText);
+                return Err(SmtpReplyError::InvalidReplyText);
             }
 
             let line = line.into_c_string();
 
             if line.as_bytes().len() > 980 {
-                return Err(ContextError::InvalidReplyText);
+                return Err(SmtpReplyError::InvalidReplyText);
             }
             if line.as_bytes().iter().any(|&b| matches!(b, b'\r' | b'\n')) {
-                return Err(ContextError::InvalidReplyText);
+                return Err(SmtpReplyError::InvalidReplyText);
             }
 
             message_lines.push(line);
@@ -294,20 +354,6 @@ struct ErrorReply {
 }
 
 impl ErrorReply {
-    // TODO single line vs multi line:
-    // single:  always "r "
-    //          then may have "x "
-    //          then may have "m"
-    //          => ie minimal is just "550 "
-    //          => regex:  /r (x|m|x m)?/
-    // multi: always "r x "
-    //        then may have no or multiple lines of text...
-    // in other words need API that supports following possibs:
-    // "r "
-    // "r x"
-    // "r m"
-    // "r x m"
-    // "r x m..."
     pub(crate) fn make_error_reply(&self) -> CString {
         fn fmt_codes(msg: &mut Vec<u8>, rcode: &ReplyCode, xcode: Option<&EnhancedStatusCode>) {
             write!(msg, "{} ", rcode.as_ref()).unwrap();
@@ -352,6 +398,7 @@ impl ErrorReply {
 
                 write!(msg, "{} {} ", rcode.as_ref(), xcode).unwrap();
                 msg.write_all(last_line.as_bytes()).unwrap();
+
                 msg
             }
         };
@@ -360,6 +407,7 @@ impl ErrorReply {
     }
 }
 
+/// End-of-message actions.
 pub struct EomActions {
     conn: Connection,
     available_actions: Actions,
@@ -372,13 +420,13 @@ impl ContextActions for EomActions {
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::ADDHDRS) {
+        if !self.available_actions.contains(Actions::ADD_HEADER) {
             return Err(ContextError::ActionUnavailable);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let value = value.into_c_string();
@@ -394,16 +442,16 @@ impl ContextActions for EomActions {
         name: impl IntoCString + Send + 'k,
         value: impl IntoCString + Send + 'v,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::ADDHDRS) {
+        if !self.available_actions.contains(Actions::ADD_HEADER) {
             return Err(ContextError::ActionUnavailable);
         }
         if index < 0 {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let value = value.into_c_string();
@@ -419,16 +467,16 @@ impl ContextActions for EomActions {
         index: i32,
         value: Option<impl IntoCString + Send + 'v>,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::CHGHDRS) {
+        if !self.available_actions.contains(Actions::CHANGE_HEADER) {
             return Err(ContextError::ActionUnavailable);
         }
         if index < 0 {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let name = name.into_c_string();
         if name.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let value = value.map_or_else(Default::default, |v| v.into_c_string());
@@ -443,13 +491,13 @@ impl ContextActions for EomActions {
         mail: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::CHGFROM) {
+        if !self.available_actions.contains(Actions::CHANGE_SENDER) {
             return Err(ContextError::ActionUnavailable);
         }
 
         let mail = mail.into_c_string();
         if mail.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let args = args.map(|a| a.into_c_string());
@@ -463,13 +511,13 @@ impl ContextActions for EomActions {
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::ADDRCPT) {
+        if !self.available_actions.contains(Actions::ADD_RCPT) {
             return Err(ContextError::ActionUnavailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         self.conn.write_reply(Reply::AddRcpt { rcpt }).await?;
@@ -482,13 +530,13 @@ impl ContextActions for EomActions {
         rcpt: impl IntoCString + Send + 'a,
         args: Option<impl IntoCString + Send + 'b>,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::ADDRCPT_PAR) {
+        if !self.available_actions.contains(Actions::ADD_RCPT_EXT) {
             return Err(ContextError::ActionUnavailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         let args = args.map(|a| a.into_c_string());
@@ -502,13 +550,13 @@ impl ContextActions for EomActions {
         &'cx self,
         rcpt: impl IntoCString + Send + 'a,
     ) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::DELRCPT) {
+        if !self.available_actions.contains(Actions::DELETE_RCPT) {
             return Err(ContextError::ActionUnavailable);
         }
 
         let rcpt = rcpt.into_c_string();
         if rcpt.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         self.conn.write_reply(Reply::DeleteRcpt { rcpt }).await?;
@@ -517,22 +565,22 @@ impl ContextActions for EomActions {
     }
 
     async fn replace_body<'cx, 'a>(&'cx self, chunk: &'a [u8]) -> Result<(), ContextError> {
-        if !self.available_actions.contains(Actions::CHGBODY) {
+        if !self.available_actions.contains(Actions::REPLACE_BODY) {
             return Err(ContextError::ActionUnavailable);
         }
 
-        if chunk.is_empty() {
-            self.conn.write_reply(Reply::ReplaceBody { chunk: vec![] }).await?;
-
-            return Ok(());
-        }
-
         const CHUNK_SIZE: usize = 65535;
 
-        for c in chunk.chunks(CHUNK_SIZE) {
-            let chunk = c.to_vec();
+        if chunk.is_empty() {
+            let chunk = Bytes::new();
 
             self.conn.write_reply(Reply::ReplaceBody { chunk }).await?;
+        } else {
+            for chunk in chunk.chunks(CHUNK_SIZE) {
+                let chunk = Bytes::copy_from_slice(chunk);
+
+                self.conn.write_reply(Reply::ReplaceBody { chunk }).await?;
+            }
         }
 
         Ok(())
@@ -554,7 +602,7 @@ impl ContextActions for EomActions {
 
         let reason = reason.into_c_string();
         if reason.as_bytes().is_empty() {
-            return Err(ContextError::InvalidReply);
+            return Err(ContextError::InvalidParams);
         }
 
         self.conn.write_reply(Reply::Quarantine { reason }).await?;
@@ -661,27 +709,55 @@ mod tests {
 
     #[test]
     fn error_reply_ok() {
-        let e = ErrorReply {
+        let reply = ErrorReply {
+            rcode: "550".parse().unwrap(),
+            xcode: None,
+            message: vec![],
+        };
+        assert_eq!(reply.make_error_reply().as_ref(), c_str!("550 "));
+
+        let reply = ErrorReply {
             rcode: "550".parse().unwrap(),
             xcode: Some("5.0.0".parse().unwrap()),
-            message: vec![c_str!("fail").into()],
+            message: vec![],
         };
+        assert_eq!(reply.make_error_reply().as_ref(), c_str!("550 5.0.0"));
 
-        let s = e.make_error_reply();
+        let reply = ErrorReply {
+            rcode: "550".parse().unwrap(),
+            xcode: None,
+            message: vec![c_str!("failure").into()],
+        };
+        assert_eq!(reply.make_error_reply().as_ref(), c_str!("550 failure"));
 
-        assert_eq!(s, c_str!("550 5.0.0 fail").into());
+        let reply = ErrorReply {
+            rcode: "550".parse().unwrap(),
+            xcode: Some("5.0.0".parse().unwrap()),
+            message: vec![c_str!("failure").into()],
+        };
+        assert_eq!(reply.make_error_reply().as_ref(), c_str!("550 5.0.0 failure"));
     }
 
     #[test]
     fn error_reply_multi_ok() {
-        let e = ErrorReply {
-            rcode: "550".parse().unwrap(),
-            xcode: Some("5.0.0".parse().unwrap()),
-            message: vec![c_str!("fail").into(), c_str!("totally").into()],
+        let reply = ErrorReply {
+            rcode: "400".parse().unwrap(),
+            xcode: None,
+            message: vec![c_str!("complete").into(), c_str!("failure").into()],
         };
+        assert_eq!(
+            reply.make_error_reply().as_ref(),
+            c_str!("400-4.0.0 complete\r\n400 4.0.0 failure")
+        );
 
-        let s = e.make_error_reply();
-
-        assert_eq!(s, c_str!("550-5.0.0 fail\r\n550 5.0.0 totally").into());
+        let reply = ErrorReply {
+            rcode: "411".parse().unwrap(),
+            xcode: Some("4.1.1".parse().unwrap()),
+            message: vec![c_str!("complete").into(), c_str!("failure").into()],
+        };
+        assert_eq!(
+            reply.make_error_reply().as_ref(),
+            c_str!("411-4.1.1 complete\r\n411 4.1.1 failure")
+        );
     }
 }
