@@ -15,13 +15,12 @@ mod proto_util;
 mod session;
 mod stream;
 
-use crate::session::Session;
 pub use crate::{
     callbacks::{CallbackFuture, Callbacks, Status},
     config::Config,
     context::{
         Context, ContextActions, ContextError, EomActions, EomContext, NegotiateContext,
-        SetErrorReply, SmtpReply,
+        SetErrorReply, SmtpReply, SmtpReplyError,
     },
     ffi_util::IntoCString,
     listener::{IntoListener, Listener},
@@ -30,40 +29,43 @@ pub use crate::{
 };
 use std::{future::Future, io, sync::Arc};
 use tokio::{
-    io::{AsyncRead, AsyncWrite},
     select,
-    sync::{watch, OwnedSemaphorePermit, Semaphore},
+    sync::{watch, Semaphore},
 };
-use tracing::{debug, error, trace};
+use tracing::{error, trace};
 
 // TODO Tracing levels policy:
 // When this library encounters an unanticipated failure condition (programming
 // error) it panics. No error logging is done in such a case.
 //
 // For all other error conditions the following policy is used.
-// As a library, the general principle is not to log about library operation
-// above debug level.
+// This being a library, the general principle is not to log about library
+// operation above debug level. In fact, this library only does logging at
+// `trace` level, except for the following cases:
 //
-// `error!`: the milter library fails to provide service, eg when no new
-// connections can be accepted due to I/O problem. Needs administrator attention
-// `warn!`: when the user-provided milter is detected to misbehave (user error),
-// eg when `Noreply` status is returned but was not negotiated beforehand
-// `info!`: not used
-// `debug!`: events related to milter library operation of more general interest
-// `trace!`: fine-grained operation events
-// TODO or perhaps *only* `trace`?
+// `error`: The milter library fails to provide service, eg when no new
+//   connections can be accepted due to an I/O problem.
+// `warn`: The user-provided milter implementation misbehaves (user error), eg
+//   when `Noreply` status is returned but was not negotiated beforehand.
 
-/// Runs a milter task that handles MTA connections until it is shut down.
+/// Runs a milter that handles MTA connections until it is shut down.
+///
+/// While the future returned by `run` is awaited, it perpetually accepts new
+/// MTA connections and spawns a session task for each connection. This
+/// procedure continues and the future will not complete until the supplied
+/// `shutdown` future completes.
 ///
 /// # Cancellation
 ///
-/// For graceful termination, the milter should be shut down by letting the
-/// `shutdown` future complete. If the `Future` returned by this function is
-/// simply dropped, currently active, spawned sessions may not exit immediately.
+/// For graceful termination, the milter task should be shut down by letting the
+/// `shutdown` future complete. If instead the future returned by `run` is
+/// simply dropped, currently active, spawned sessions may continue to execute.
 ///
 /// # Errors
 ///
-/// To do.
+/// When the given `listener` cannot be converted into a [`Listener`] an error
+/// is returned. When the listener fails to accept any new connections an error
+/// is returned. Else, the task runs for ever until it is shut down.
 ///
 /// # Example
 ///
@@ -92,8 +94,6 @@ pub async fn run<T>(
 where
     T: Send + 'static,
 {
-    debug!("milter starting");
-
     let listener = listener.into_listener()?;
 
     // The supplied shutdown_milter causes the main connection/session spawn in
@@ -104,20 +104,23 @@ where
     let shutdown_milter = shutdown;
     let (shutdown_sessions, _) = watch::channel(false);
 
+    trace!("milter starting");
+
     // The invocation of `run_milter` never returns normally. It has an infinite
     // loop that is only broken when the listener cannot accept any new
     // connections.
     // When the shutdown future completes, the `run_milter` future is simply
-    // dropped in the middle of whatever it is doing.
+    // dropped in the middle of whatever it is doing. The listener is dropped at
+    // the same time and no new connections are accepted.
 
     let result = select! {
-        res = run_milter(&listener, callbacks, config, &shutdown_sessions) => {
+        res = run_milter(listener, callbacks, config, &shutdown_sessions) => {
             let e = res.unwrap_err();
             error!("milter exited with error, shutting down: {}", e);
             Err(e)
         }
         _ = shutdown_milter => {
-            debug!("milter shutting down");
+            trace!("milter shutting down");
             Ok(())
         }
     };
@@ -131,34 +134,32 @@ where
     result
 }
 
-// Main loop spawning session tasks, that handle commands coming in on a
-// session. Runs for ever.
-
-// TODO However, there are internal error conditions that *should* result in
-// loop exit and Err result: for example if the listener somehow breaks and
-// cannot be used to obtain connections any more. In that case the fault is not
-// of some individual connection (which is not treated as an error here), but
-// ours, and should be propagated.
-
+// Main loop spawning session tasks that handle commands coming in on a session.
+// Normally runs for ever.
+//
+// However, there are internal error conditions that *should* result in loop
+// exit and Err result: for example if the listener somehow breaks and cannot be
+// used to obtain connections any more. In that case the fault is not of some
+// individual connection (which is not treated as an error here), but ours, and
+// should be propagated.
+//
 // Arguments are moved into the `run_milter` future, so that when it is dropped,
 // associated resources (except the ones only borrowed) are dropped at the same
 // time, too.
-async fn run_milter<T: Send + 'static>(
-    listener: &Listener,
+async fn run_milter<T>(
+    listener: Listener,
     callbacks: Callbacks<T>,
     config: Config,
     shutdown_sender: &watch::Sender<bool>,
-) -> io::Result<()> {
+) -> io::Result<()>
+where
+    T: Send + 'static,
+{
     let callbacks = Arc::new(callbacks);
     let config = Arc::new(config);
 
     let conn_permits = Arc::new(Semaphore::new(config.max_connections));
 
-    // This is a diverging function, ie the following loop runs for ever. Since
-    // this function is called in a `select!` together with the shutdown signal,
-    // the future produced by this function including all captured data will
-    // simply be dropped on shutdown (at one of the following await points).
-
     loop {
         // Spawn new sessions continuously, but make sure that no more than the
         // max connections limit are in flight at the same time.
@@ -169,57 +170,17 @@ async fn run_milter<T: Send + 'static>(
         // Wait for a connection. Then, ready to go: synchronously spawn a new
         // session to tokio and resume looping.
 
-        match listener {
+        match &listener {
             Listener::Tcp(listener) => {
                 let (stream, _) = listener.accept().await?;
 
-                spawn_session(stream, shutdown_sender, &callbacks, &config, permit);
+                session::spawn(stream, shutdown_sender, &callbacks, &config, permit);
             }
             Listener::Unix(listener) => {
                 let (stream, _) = listener.accept().await?;
 
-                spawn_session(stream, shutdown_sender, &callbacks, &config, permit);
+                session::spawn(stream, shutdown_sender, &callbacks, &config, permit);
             }
         };
     }
 }
-
-fn spawn_session<S, T>(
-    stream: S,
-    shutdown_sender: &watch::Sender<bool>,
-    callbacks: &Arc<Callbacks<T>>,
-    config: &Config,
-    permit: OwnedSemaphorePermit,
-) where
-    S: AsyncRead + AsyncWrite + Unpin + Send + 'static,
-    T: Send + 'static,
-{
-    let callbacks = callbacks.clone();
-
-    let shutdown = shutdown_sender.subscribe();
-
-    let session = Session::new(stream, shutdown, callbacks, config);
-
-    tokio::spawn(async move {
-        trace!("session beginning processing commands");
-
-        // Every sessions runs until it either exits regularly (eg client issues
-        // the "quit" command), or it exits with an error. An error is only
-        // logged but not otherwise propagated.
-
-        match session.process_commands().await {
-            Ok(()) => {
-                trace!("session done processing commands");
-            }
-            Err(e) => {
-                trace!("error in session while processing commands: {}", e);
-            }
-        }
-
-        // At this point the session and the stream that it contains are already
-        // dropped. The final thing to do is to drop the permit to allow another
-        // client to connect.
-
-        drop(permit);
-    });
-}
