@@ -1,10 +1,11 @@
-use crate::message::commands::CommandError;
+use crate::message::{TryFromByteError, TryFromIndexError};
 use std::{
     collections::HashMap,
     ffi::{CStr, CString},
 };
 
-#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
+/// The milter stage.
+#[derive(Clone, Copy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub enum Stage {
     Connect,
     Helo,
@@ -16,10 +17,17 @@ pub enum Stage {
 }
 
 impl Stage {
+    /// Returns an iterator of all stages.
     pub fn all_stages() -> impl DoubleEndedIterator<Item = Self> {
         use Stage::*;
         [Connect, Helo, Mail, Rcpt, Data, Eoh, Eom].into_iter()
     }
+
+    /// Returns an iterator of all stages, sorted by their index value.
+    pub fn all_stages_sorted_by_index() -> impl DoubleEndedIterator<Item = Self> {
+        use Stage::*;
+        [Connect, Helo, Mail, Rcpt, Data, Eom, Eoh].into_iter()
+    }
 }
 
 impl From<Stage> for i32 {
@@ -30,8 +38,25 @@ impl From<Stage> for i32 {
             Stage::Mail => 2,
             Stage::Rcpt => 3,
             Stage::Data => 4,
-            Stage::Eoh => 5,
-            Stage::Eom => 6,
+            Stage::Eoh => 6,
+            Stage::Eom => 5,
+        }
+    }
+}
+
+impl TryFrom<i32> for Stage {
+    type Error = TryFromIndexError;
+
+    fn try_from(value: i32) -> Result<Self, Self::Error> {
+        match value {
+            0 => Ok(Self::Connect),
+            1 => Ok(Self::Helo),
+            2 => Ok(Self::Mail),
+            3 => Ok(Self::Rcpt),
+            4 => Ok(Self::Data),
+            6 => Ok(Self::Eoh),
+            5 => Ok(Self::Eom),
+            value => Err(TryFromIndexError::new(value)),
         }
     }
 }
@@ -50,27 +75,8 @@ impl From<Stage> for u8 {
     }
 }
 
-impl TryFrom<i32> for Stage {
-    // TODO not CommandError
-    type Error = CommandError;
-
-    fn try_from(value: i32) -> Result<Self, Self::Error> {
-        match value {
-            0 => Ok(Self::Connect),
-            1 => Ok(Self::Helo),
-            2 => Ok(Self::Mail),
-            3 => Ok(Self::Rcpt),
-            4 => Ok(Self::Data),
-            5 => Ok(Self::Eoh),
-            6 => Ok(Self::Eom),
-            _ => Err(CommandError::UnknownStage),
-        }
-    }
-}
-
 impl TryFrom<u8> for Stage {
-    // TODO not CommandError
-    type Error = CommandError;
+    type Error = TryFromByteError;
 
     fn try_from(value: u8) -> Result<Self, Self::Error> {
         match value {
@@ -81,13 +87,13 @@ impl TryFrom<u8> for Stage {
             b'T' => Ok(Self::Data),
             b'N' => Ok(Self::Eoh),
             b'E' => Ok(Self::Eom),
-            _ => Err(CommandError::UnknownStage),
+            value => Err(TryFromByteError::new(value)),
         }
     }
 }
 
-// TODO *not* Default or Clone: this is an encapsulating type that is not meant
-// to be created outside this crate
+// Note: This struct does *not* implement `Default` or `Clone`. Callbacks
+// receive a mutable reference to this struct but must not be able to remove it.
 /// Currently defined macros.
 #[derive(Debug)]
 pub struct Macros(HashMap<Stage, HashMap<CString, CString>>);
@@ -105,36 +111,38 @@ impl Macros {
         Self(self.0.clone())
     }
 
-    pub fn get(&self, key: &CStr) -> Option<&CStr> {
-        let alt_key;
-        let alt_key = match *key.to_bytes() {
+    /// Returns the value associated with the given macro name, if any.
+    pub fn get(&self, name: &CStr) -> Option<&CStr> {
+        let alt_name;
+        let alt_name = match *name.to_bytes() {
             [x] => {
-                alt_key = [b'{', x, b'}', 0];
-                Some(CStr::from_bytes_with_nul(&alt_key[..4]).unwrap())
+                alt_name = [b'{', x, b'}', 0];
+                Some(CStr::from_bytes_with_nul(&alt_name[..4]).unwrap())
             }
             [b'{', x, b'}'] => {
-                alt_key = [x, 0, 0, 0];
-                Some(CStr::from_bytes_with_nul(&alt_key[..2]).unwrap())
+                alt_name = [x, 0, 0, 0];
+                Some(CStr::from_bytes_with_nul(&alt_name[..2]).unwrap())
             }
             _ => None,
         };
 
         Stage::all_stages().rev().find_map(|stage| {
             let m = &self.0[&stage];
-            m.get(key)
-                .or_else(|| alt_key.and_then(|k| m.get(k)))
+            m.get(name)
+                .or_else(|| alt_name.and_then(|name| m.get(name)))
                 .map(|v| v.as_ref())
         })
     }
 
+    /// Returns a new hash map with all defined macros.
     pub fn to_hash_map(&self) -> HashMap<CString, CString> {
         Stage::all_stages()
             .flat_map(|stage| self.0[&stage].clone())
             .collect()
     }
 
-    pub(crate) fn insert(&mut self, key: Stage, entries: HashMap<CString, CString>) {
-        self.0.insert(key, entries);
+    pub(crate) fn insert(&mut self, stage: Stage, entries: HashMap<CString, CString>) {
+        self.0.insert(stage, entries);
     }
 
     pub(crate) fn clear(&mut self) {
@@ -176,4 +184,43 @@ mod tests {
         assert_eq!(macros.get(c_str!("x")), Some(c_str!("y")));
         assert_eq!(macros.get(c_str!("{x}")), Some(c_str!("y")));
     }
+
+    #[test]
+    fn to_hash_map_ok() {
+        let mut macros = Macros::new();
+
+        macros.insert(
+            Stage::Connect,
+            HashMap::from([
+                (c_str!("{name1}").into(), c_str!("connect1").into()),
+                (c_str!("{name2}").into(), c_str!("connect2").into()),
+            ]),
+        );
+        macros.insert(
+            Stage::Mail,
+            HashMap::from([
+                (c_str!("{name2}").into(), c_str!("mail2").into()),
+                (c_str!("{name3}").into(), c_str!("mail3").into()),
+            ]),
+        );
+
+        assert_eq!(
+            macros.to_hash_map(),
+            HashMap::from([
+                (c_str!("{name1}").into(), c_str!("connect1").into()),
+                (c_str!("{name2}").into(), c_str!("mail2").into()),
+                (c_str!("{name3}").into(), c_str!("mail3").into()),
+            ]),
+        );
+
+        macros.clear_after(Stage::Helo);
+
+        assert_eq!(
+            macros.to_hash_map(),
+            HashMap::from([
+                (c_str!("{name1}").into(), c_str!("connect1").into()),
+                (c_str!("{name2}").into(), c_str!("connect2").into()),
+            ]),
+        );
+    }
 }
