@@ -6,12 +6,11 @@ use crate::{
     macros::Stage,
     message::{
         commands::{
-            BodyPayload, CommandError, CommandKind, CommandMessage, ConnInfoPayload,
-            EnvAddrPayload, HeaderPayload, HeloPayload, MacroPayload, OptNegPayload,
-            UnknownPayload,
+            CommandKind, CommandMessage, ConnInfoPayload, EnvAddrPayload, HeaderPayload,
+            HeloPayload, MacroPayload, OptNegPayload, ParseCommandError, UnknownPayload,
         },
         replies::Reply,
-        Version, MILTER_VERSION,
+        Version, PROTOCOL_VERSION,
     },
     proto_util::{Actions, ProtoOpts},
 };
@@ -27,9 +26,9 @@ use std::{
 use tokio::{
     io::{AsyncRead, AsyncWrite},
     select,
-    sync::watch,
+    sync::{watch, OwnedSemaphorePermit},
 };
-use tracing::{debug, trace, warn};
+use tracing::{trace, warn};
 
 #[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
 pub enum State {
@@ -67,15 +66,15 @@ impl State {
 
     fn can_reach(&self, target: Self, opts: ProtoOpts) -> bool {
         let can_be_skipped = |s: &Self| match s {
-            Self::Conn => opts.contains(ProtoOpts::NOCONNECT),
-            Self::Helo => opts.contains(ProtoOpts::NOHELO),
-            Self::Mail => opts.contains(ProtoOpts::NOMAIL),
-            Self::Rcpt => opts.contains(ProtoOpts::NORCPT),
-            Self::Header => opts.contains(ProtoOpts::NOHDRS),
-            Self::Eoh => opts.contains(ProtoOpts::NOEOH),
-            Self::Body => opts.contains(ProtoOpts::NOBODY),
-            Self::Data => opts.contains(ProtoOpts::NODATA),
-            Self::Unknown => opts.contains(ProtoOpts::NOUNKNOWN),
+            Self::Conn => opts.contains(ProtoOpts::NO_CONNECT),
+            Self::Helo => opts.contains(ProtoOpts::NO_HELO),
+            Self::Mail => opts.contains(ProtoOpts::NO_MAIL),
+            Self::Rcpt => opts.contains(ProtoOpts::NO_RCPT),
+            Self::Data => opts.contains(ProtoOpts::NO_DATA),
+            Self::Header => opts.contains(ProtoOpts::NO_HEADER),
+            Self::Eoh => opts.contains(ProtoOpts::NO_EOH),
+            Self::Body => opts.contains(ProtoOpts::NO_BODY),
+            Self::Unknown => opts.contains(ProtoOpts::NO_UNKNOWN),
             _ => false,
         };
 
@@ -100,10 +99,7 @@ impl State {
             Opts | QuitNc => matches!(next, Conn | Unknown),
             Conn | Helo => matches!(next, Helo | Mail | Unknown),
             Mail => matches!(next, Rcpt | Abort | Unknown),
-            Rcpt => matches!(
-                next,
-                Header | Eoh | Data | Body | Eom | Rcpt | Abort | Unknown
-            ),
+            Rcpt => matches!(next, Header | Eoh | Data | Body | Eom | Rcpt | Abort | Unknown),
             Data | Header => matches!(next, Eoh | Header | Abort),
             Eoh | Body => matches!(next, Body | Eom | Abort),
             Eom => matches!(next, Quit | Mail | Unknown | QuitNc),
@@ -122,39 +118,54 @@ impl State {
 
 fn opts_from_callbacks<T: Send>(callbacks: &Callbacks<T>) -> ProtoOpts {
     let mut opts = ProtoOpts::empty();
-    opts.set(ProtoOpts::NOCONNECT, callbacks.connect.is_none());
-    opts.set(ProtoOpts::NOHELO, callbacks.helo.is_none());
-    opts.set(ProtoOpts::NOMAIL, callbacks.mail.is_none());
-    opts.set(ProtoOpts::NORCPT, callbacks.rcpt.is_none());
-    opts.set(ProtoOpts::NOHDRS, callbacks.header.is_none());
-    opts.set(ProtoOpts::NOEOH, callbacks.eoh.is_none());
-    opts.set(ProtoOpts::NOBODY, callbacks.body.is_none());
-    opts.set(ProtoOpts::NODATA, callbacks.data.is_none());
-    opts.set(ProtoOpts::NOUNKNOWN, callbacks.unknown.is_none());
+    opts.set(ProtoOpts::NO_CONNECT, callbacks.connect.is_none());
+    opts.set(ProtoOpts::NO_HELO, callbacks.helo.is_none());
+    opts.set(ProtoOpts::NO_MAIL, callbacks.mail.is_none());
+    opts.set(ProtoOpts::NO_RCPT, callbacks.rcpt.is_none());
+    opts.set(ProtoOpts::NO_DATA, callbacks.data.is_none());
+    opts.set(ProtoOpts::NO_HEADER, callbacks.header.is_none());
+    opts.set(ProtoOpts::NO_EOH, callbacks.eoh.is_none());
+    opts.set(ProtoOpts::NO_BODY, callbacks.body.is_none());
+    opts.set(ProtoOpts::NO_UNKNOWN, callbacks.unknown.is_none());
     opts
 }
 
-pub type SessionResult = Result<(), SessionError>;
+type SessionResult = Result<(), SessionError>;
 
-// Reasons for abnormal session termination.
+/// Reasons for abnormal session termination.
 #[derive(Debug)]
-pub enum SessionError {
+enum SessionError {
     Io(io::Error),
-    ParseCommand(CommandError),
+    ParseCommand(ParseCommandError),
     ProtocolVersionNotSupported,
-    RequestedActionsNotSupported, // requested actions cannot be fulfilled
-    RequestedProtoOptsNotSupported,
+    ActionsNotSupported,
+    ProtoOptsNotSupported,
     InvalidNegotiateStatus(Status),
-    BufferEmpty,  // empty buffer not allowed
-    UnknownCommand,
-    Shutdown,
-    // TODO...
-    // TODO timeout?
+    BufferEmpty(CommandKind),
+    UnknownCommand(u8),
+    MilterShutDown,
 }
 
+// TODO human readable display for debugging
 impl Display for SessionError {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        write!(f, "{:?}", self)
+        match self {
+            Self::Io(error) => write!(f, "I/O error: {}", error),
+            Self::ParseCommand(error) => write!(f, "command could not be parsed: {}", error),
+            Self::ProtocolVersionNotSupported => {
+                write!(f, "requested milter protocol version not supported")
+            }
+            Self::ActionsNotSupported => write!(f, "requested actions not supported"),
+            Self::ProtoOptsNotSupported => {
+                write!(f, "requested milter protocol options not supported")
+            }
+            Self::InvalidNegotiateStatus(status) => {
+                write!(f, "invalid status in negotiation: {:?}", status)
+            }
+            Self::BufferEmpty(cmd) => write!(f, "command with empty payload buffer: {:?}", cmd),
+            Self::UnknownCommand(byte) => write!(f, "unknown command: {:?}", byte),
+            Self::MilterShutDown => write!(f, "milter shut down"),
+        }
     }
 }
 
@@ -166,13 +177,61 @@ impl From<io::Error> for SessionError {
     }
 }
 
-impl From<CommandError> for SessionError {
-    fn from(error: CommandError) -> Self {
+impl From<ParseCommandError> for SessionError {
+    fn from(error: ParseCommandError) -> Self {
         Self::ParseCommand(error)
     }
 }
 
-pub struct Session<T: Send> {
+/// Synchronously spawns a new session.
+pub fn spawn<S, T>(
+    stream: S,
+    shutdown_sender: &watch::Sender<bool>,
+    callbacks: &Arc<Callbacks<T>>,
+    config: &Config,
+    permit: OwnedSemaphorePermit,
+) where
+    S: AsyncRead + AsyncWrite + Unpin + Send + 'static,
+    T: Send + 'static,
+{
+    let callbacks = callbacks.clone();
+
+    let shutdown = shutdown_sender.subscribe();
+
+    let session = Session::new(stream, shutdown, callbacks, config);
+
+    tokio::spawn(async move {
+        trace!("session beginning processing commands");
+
+        // Every sessions runs until it either exits regularly (eg client issues
+        // the Quit command), or it exits with an error. An error is only logged
+        // but not otherwise propagated.
+
+        match session.process_commands().await {
+            Ok(()) => {
+                trace!("session done processing commands");
+            }
+            Err(e) => {
+                trace!("error in session while processing commands: {}", e);
+            }
+        }
+
+        // At this point the session and the stream that it contains are already
+        // dropped. The final thing to do is to drop the permit to allow another
+        // client to connect.
+
+        drop(permit);
+    });
+}
+
+// It is unclear what to do with the QuitNc command. It seems to me that
+// libmilter does not implement this command properly. On QuitNc, a session can
+// again reenter the negotiate stage; libmilter stays in the main loop, but
+// without resetting the connection info: this would leak context data across
+// connections. We do the same as libmilter (and assume that no MTA actually
+// uses QuitNc …).
+
+struct Session<T: Send> {
     conn: Connection,
 
     state: State,
@@ -186,11 +245,8 @@ pub struct Session<T: Send> {
     opts: ProtoOpts,
 }
 
-// TODO initialization: note that with QuitNc command, this session can again
-// begin at the negotiate stage: ie need to make sure all fields are reset to
-// initial values?
 impl<T: Send> Session<T> {
-    pub fn new<S>(
+    fn new<S>(
         stream: S,
         shutdown: watch::Receiver<bool>,
         callbacks: Arc<Callbacks<T>>,
@@ -222,8 +278,8 @@ impl<T: Send> Session<T> {
     // This takes session by value, as it is called only once. Therefore, when
     // the invocation returns in the task that calls it, the session and
     // associated values are already dropped.
-    pub async fn process_commands(mut self) -> SessionResult {
-        let result = self.process().await;
+    async fn process_commands(mut self) -> SessionResult {
+        let result = self.process_until_done().await;
 
         if result.is_err() && self.state.is_mail_transaction() {
             if let Some(abort) = &self.callbacks.abort {
@@ -244,13 +300,13 @@ impl<T: Send> Session<T> {
         result
     }
 
-    async fn process(&mut self) -> SessionResult {
+    async fn process_until_done(&mut self) -> SessionResult {
         // For ever process this session’s connection by handling commands until
         // the quit command is received, or the shutdown signal is received.
 
         while self.state != State::Quit {
             if *self.shutdown.borrow() {
-                return Err(SessionError::Shutdown);
+                return Err(SessionError::MilterShutDown);
             }
 
             // Read the next message from the connection. Always also check if
@@ -267,7 +323,7 @@ impl<T: Send> Session<T> {
             // certain that a callback will be called.
 
             let msg = CommandMessage::try_from(msg)
-                .map_err(|_| SessionError::UnknownCommand)?;
+                .map_err(|e| SessionError::UnknownCommand(e.byte()))?;
 
             let cmd = msg.kind;
 
@@ -289,14 +345,14 @@ impl<T: Send> Session<T> {
                     self.state = State::Helo;
 
                     if !self.state.can_reach(next_state, self.opts) {
-                        // The desired state cannot be reached; but if it's a
+                        // The desired state cannot be reached; but if it is a
                         // quit anyway then exit regularly. Else, *ignore* the
                         // invalid command.
 
                         if next_state == State::Quit {
                             break;
                         } else {
-                            debug!("ignoring unexpected command");
+                            trace!("ignoring unexpected command");
                             continue;
                         }
                     }
@@ -404,7 +460,7 @@ impl<T: Send> Session<T> {
             }
             CommandKind::BodyChunk => {
                 if let Some(body) = &self.callbacks.body {
-                    let BodyPayload { chunk } = BodyPayload::parse_buffer(msg.buffer)?;
+                    let chunk = msg.buffer.as_ref().into();
                     body(&mut self.context, chunk).await
                 } else {
                     Status::Continue
@@ -415,9 +471,12 @@ impl<T: Send> Session<T> {
 
                 if let Some(body) = &self.callbacks.body {
                     if !msg.buffer.is_empty() {
-                        let BodyPayload { chunk } = BodyPayload::parse_buffer(msg.buffer)?;
+                        let chunk = msg.buffer.as_ref().into();
                         status = body(&mut self.context, chunk).await;
 
+                        // This logic is suspicious, but it is just like in
+                        // libmilter: If a reply is written here, nevertheless a
+                        // (second!) reply will be written further below.
                         if status != Status::Continue {
                             self.write_reply(status).await?;
                         }
@@ -484,8 +543,8 @@ impl<T: Send> Session<T> {
     }
 
     async fn handle_opt_neg_command(&mut self, buffer: Bytes) -> SessionResult {
-        // At this stage there are usually no macros to be cleared, however,
-        // note that macros can be sent at any time, even before negotiate.
+        // At this stage there are usually no macros to be cleared. However,
+        // macros could be sent at any time, even before negotiate.
         self.context.clear_macros();
 
         let OptNegPayload {
@@ -494,21 +553,19 @@ impl<T: Send> Session<T> {
             opts: mut mta_opts,
         } = OptNegPayload::parse_buffer(buffer)?;
 
-        // Note: different from libmilter. We only support protocol versions 6
-        // (and up). Older versions are not supported.
-        if mta_version < MILTER_VERSION {
+        // Note: this is different from libmilter. We only support protocol
+        // versions 6 (and up). Older versions are not supported.
+        if mta_version < PROTOCOL_VERSION {
             return Err(SessionError::ProtocolVersionNotSupported);
         }
 
         if mta_actions.is_empty() {
-            mta_actions = Actions::MIN_ACTIONS;
+            mta_actions = Actions::min_actions();
         }
         if mta_opts.is_empty() {
-            mta_opts = ProtoOpts::MIN_PROTO_OPTS;
+            mta_opts = ProtoOpts::min_proto_opts();
         }
 
-        // TODO FFR max data size?
-
         let target_opts;
         let requested_macros;
 
@@ -552,14 +609,19 @@ impl<T: Send> Session<T> {
         }
 
         if !mta_actions.contains(self.actions) {
-            return Err(SessionError::RequestedActionsNotSupported);
+            return Err(SessionError::ActionsNotSupported);
         }
         if !mta_opts.contains(target_opts) {
-            return Err(SessionError::RequestedProtoOptsNotSupported);
+            return Err(SessionError::ProtoOptsNotSupported);
         }
 
-        self.write_opt_neg_reply(MILTER_VERSION, self.actions, target_opts, requested_macros)
-            .await?;
+        self.write_opt_neg_reply(
+            PROTOCOL_VERSION,
+            self.actions,
+            target_opts,
+            requested_macros,
+        )
+        .await?;
 
         Ok(())
     }
@@ -617,15 +679,15 @@ impl<T: Send> Session<T> {
     async fn write_reply(&mut self, mut status: Status) -> io::Result<()> {
         fn needs_noreply(opts: ProtoOpts, state: State) -> bool {
             match state {
-                State::Conn => opts.contains(ProtoOpts::NR_CONN),
-                State::Helo => opts.contains(ProtoOpts::NR_HELO),
-                State::Mail => opts.contains(ProtoOpts::NR_MAIL),
-                State::Rcpt => opts.contains(ProtoOpts::NR_RCPT),
-                State::Data => opts.contains(ProtoOpts::NR_DATA),
-                State::Unknown => opts.contains(ProtoOpts::NR_UNKN),
-                State::Header => opts.contains(ProtoOpts::NR_HDR),
-                State::Eoh => opts.contains(ProtoOpts::NR_EOH),
-                State::Body => opts.contains(ProtoOpts::NR_BODY),
+                State::Conn => opts.contains(ProtoOpts::NOREPLY_CONNECT),
+                State::Helo => opts.contains(ProtoOpts::NOREPLY_HELO),
+                State::Mail => opts.contains(ProtoOpts::NOREPLY_MAIL),
+                State::Rcpt => opts.contains(ProtoOpts::NOREPLY_RCPT),
+                State::Data => opts.contains(ProtoOpts::NOREPLY_DATA),
+                State::Header => opts.contains(ProtoOpts::NOREPLY_HEADER),
+                State::Eoh => opts.contains(ProtoOpts::NOREPLY_EOH),
+                State::Body => opts.contains(ProtoOpts::NOREPLY_BODY),
+                State::Unknown => opts.contains(ProtoOpts::NOREPLY_UNKNOWN),
                 _ => false,
             }
         }
@@ -671,7 +733,7 @@ fn ensure_buffer_present(cmd: CommandKind, buf: &[u8]) -> SessionResult {
         DefMacros | BodyChunk | ConnInfo | Helo | Header | Mail | OptNeg | Rcpt | Unknown
     ) && buf.is_empty()
     {
-        return Err(SessionError::BufferEmpty);
+        return Err(SessionError::BufferEmpty(cmd));
     }
 
     Ok(())
@@ -685,13 +747,13 @@ mod tests {
     fn can_reach_ok() {
         use State::*;
 
-        let opts = ProtoOpts::NOHELO
-            | ProtoOpts::NOMAIL
-            | ProtoOpts::NOHDRS
-            | ProtoOpts::NOEOH
-            | ProtoOpts::NOBODY
-            | ProtoOpts::NODATA
-            | ProtoOpts::NOUNKNOWN;
+        let opts = ProtoOpts::NO_HELO
+            | ProtoOpts::NO_MAIL
+            | ProtoOpts::NO_DATA
+            | ProtoOpts::NO_HEADER
+            | ProtoOpts::NO_EOH
+            | ProtoOpts::NO_BODY
+            | ProtoOpts::NO_UNKNOWN;
 
         assert!(Init.can_reach(Opts, opts));
         assert!(Opts.can_reach(Conn, opts));
