@@ -2,20 +2,21 @@ mod common;
 
 pub use crate::common::*;
 use byte_strings::c_str;
+use bytes::Bytes;
 use indymilter::{
     message::{
         commands::{
-            BodyPayload, Command, ConnInfoPayload, EnvAddrPayload, HeaderPayload, HeloPayload,
-            OptNegPayload,
+            Command, ConnInfoPayload, EnvAddrPayload, HeaderPayload, HeloPayload, OptNegPayload,
         },
         replies::Reply,
-        MILTER_VERSION,
+        PROTOCOL_VERSION,
     },
-    Actions, Callbacks, ProtoOpts, Status,
+    Actions, Callbacks, ProtoOpts, SocketInfo, Status,
 };
 use std::{
     collections::HashMap,
     io::ErrorKind,
+    net::Ipv4Addr,
     sync::{Arc, Mutex},
     time::Duration,
 };
@@ -23,7 +24,7 @@ use tokio::time;
 
 #[tokio::test]
 async fn basic1() {
-    let _ = tracing_subscriber::fmt::try_init();
+    init_tracing_subscriber();
 
     let milter = Milter::spawn(LOCALHOST, Callbacks::<()>::new(), Default::default())
         .await
@@ -33,7 +34,7 @@ async fn basic1() {
 
     client
         .write_command(Command::OptNeg(OptNegPayload {
-            version: MILTER_VERSION,
+            version: PROTOCOL_VERSION,
             actions: Actions::all(),
             opts: ProtoOpts::all(),
         }))
@@ -44,17 +45,17 @@ async fn basic1() {
     assert_eq!(
         reply,
         Reply::OptNeg {
-            version: MILTER_VERSION,
+            version: PROTOCOL_VERSION,
             actions: Actions::empty(),
-            opts: ProtoOpts::NOCONNECT
-                | ProtoOpts::NOHELO
-                | ProtoOpts::NOMAIL
-                | ProtoOpts::NORCPT
-                | ProtoOpts::NOBODY
-                | ProtoOpts::NOHDRS
-                | ProtoOpts::NOEOH
-                | ProtoOpts::NOUNKNOWN
-                | ProtoOpts::NODATA,
+            opts: ProtoOpts::NO_CONNECT
+                | ProtoOpts::NO_HELO
+                | ProtoOpts::NO_MAIL
+                | ProtoOpts::NO_RCPT
+                | ProtoOpts::NO_DATA
+                | ProtoOpts::NO_HEADER
+                | ProtoOpts::NO_EOH
+                | ProtoOpts::NO_BODY
+                | ProtoOpts::NO_UNKNOWN,
             macros: HashMap::new(),
         }
     );
@@ -68,7 +69,7 @@ async fn basic1() {
 
 #[tokio::test]
 async fn ignore_unexpected_command() {
-    let _ = tracing_subscriber::fmt::try_init();
+    init_tracing_subscriber();
 
     let milter = Milter::spawn(LOCALHOST, Callbacks::<()>::new(), Default::default())
         .await
@@ -100,7 +101,7 @@ async fn ignore_unexpected_command() {
     client
         .write_command(Command::ConnInfo(ConnInfoPayload {
             hostname: c_str!("mail.example.com").into(),
-            socket_info: None,
+            socket_info: SocketInfo::Unknown,
         }))
         .await
         .unwrap();
@@ -122,7 +123,7 @@ async fn ignore_unexpected_command() {
 
 #[tokio::test]
 async fn invalid_command() {
-    let _ = tracing_subscriber::fmt::try_init();
+    init_tracing_subscriber();
 
     let milter = Milter::spawn(LOCALHOST, Callbacks::<()>::new(), Default::default())
         .await
@@ -142,11 +143,16 @@ async fn invalid_command() {
 
 #[tokio::test]
 async fn all_stages() {
-    let _ = tracing_subscriber::fmt::try_init();
+    init_tracing_subscriber();
 
     let result = Arc::new(Mutex::new(Vec::new()));
 
     let callbacks = {
+        fn push_stage(data: &mut Option<Vec<&'static str>>, stage: &'static str) -> Status {
+            data.as_mut().unwrap().push(stage);
+            Status::Continue
+        }
+
         let result = result.clone();
 
         Callbacks::new()
@@ -188,7 +194,7 @@ async fn all_stages() {
 
     client
         .write_command(Command::OptNeg(OptNegPayload {
-            version: MILTER_VERSION,
+            version: PROTOCOL_VERSION,
             actions: Actions::all(),
             opts: ProtoOpts::all(),
         }))
@@ -201,7 +207,7 @@ async fn all_stages() {
     client
         .write_command(Command::ConnInfo(ConnInfoPayload {
             hostname: c_str!("mail.example.com").into(),
-            socket_info: None,
+            socket_info: SocketInfo::Inet((Ipv4Addr::LOCALHOST, 1234).into()),
         }))
         .await
         .unwrap();
@@ -261,9 +267,7 @@ async fn all_stages() {
     assert_eq!(reply, Reply::Continue);
 
     client
-        .write_command(Command::BodyChunk(BodyPayload {
-            chunk: b"body".to_vec(),
-        }))
+        .write_command(Command::BodyChunk(Bytes::from_static(b"body")))
         .await
         .unwrap();
 
@@ -271,7 +275,7 @@ async fn all_stages() {
     assert_eq!(reply, Reply::Continue);
 
     client
-        .write_command(Command::BodyEnd(BodyPayload { chunk: vec![] }))
+        .write_command(Command::BodyEnd(Bytes::new()))
         .await
         .unwrap();
 
@@ -303,9 +307,3 @@ async fn all_stages() {
         ]
     );
 }
-
-fn push_stage(data: &mut Option<Vec<&'static str>>, stage: &'static str) -> Status {
-    data.as_mut().unwrap().push(stage);
-
-    Status::Continue
-}
